@@ -3,7 +3,7 @@
 import itertools
 from itertools import chain
 from contextlib import contextmanager
-from mypy.ordered_dict import OrderedDict
+from mypy.backports import OrderedDict
 
 from typing import Callable, List, Optional, Set, Tuple, Iterator, TypeVar, Iterable, Sequence
 from typing_extensions import Final
@@ -12,16 +12,17 @@ from mypy_extensions import DefaultNamedArg
 from mypy.messages import MessageBuilder, quote_type_string, format_type_bare
 from mypy.options import Options
 from mypy.types import (
-    Type, UnboundType, TypeVarType, TupleType, TypedDictType, UnionType, Instance, AnyType,
-    CallableType, NoneType, ErasedType, DeletedType, TypeList, TypeVarDef, SyntheticTypeVisitor,
-    StarType, PartialType, EllipsisType, UninhabitedType, TypeType,
-    CallableArgument, TypeQuery, union_items, TypeOfAny, LiteralType, RawExpressionType,
-    PlaceholderType, Overloaded, get_proper_type, TypeAliasType, TypeVarLikeDef, ParamSpecDef
+    Type, UnboundType, TupleType, TypedDictType, UnionType, Instance, AnyType,
+    CallableType, NoneType, ErasedType, DeletedType, TypeList, TypeVarType, SyntheticTypeVisitor,
+    StarType, PartialType, EllipsisType, UninhabitedType, TypeType, CallableArgument,
+    TypeQuery, union_items, TypeOfAny, LiteralType, RawExpressionType,
+    PlaceholderType, Overloaded, get_proper_type, TypeAliasType, RequiredType,
+    TypeVarLikeType, ParamSpecType, ParamSpecFlavor, callable_with_ellipsis
 )
 
 from mypy.nodes import (
     TypeInfo, Context, SymbolTableNode, Var, Expression,
-    get_nongen_builtins, check_arg_names, check_arg_kinds, ARG_POS, ARG_NAMED,
+    get_nongen_builtins, check_arg_names, check_arg_kinds, ArgKind, ARG_POS, ARG_NAMED,
     ARG_OPT, ARG_NAMED_OPT, ARG_STAR, ARG_STAR2, TypeVarExpr, TypeVarLikeExpr, ParamSpecExpr,
     TypeAlias, PlaceholderNode, SYMBOL_FUNCBASE_TYPES, Decorator, MypyFile
 )
@@ -35,7 +36,7 @@ from mypy import nodes, message_registry, errorcodes as codes
 
 T = TypeVar('T')
 
-type_constructors = {
+type_constructors: Final = {
     'typing.Callable',
     'typing.Optional',
     'typing.Tuple',
@@ -45,22 +46,22 @@ type_constructors = {
     'typing_extensions.Literal',
     'typing.Annotated',
     'typing_extensions.Annotated',
-}  # type: Final
+}
 
-ARG_KINDS_BY_CONSTRUCTOR = {
+ARG_KINDS_BY_CONSTRUCTOR: Final = {
     'mypy_extensions.Arg': ARG_POS,
     'mypy_extensions.DefaultArg': ARG_OPT,
     'mypy_extensions.NamedArg': ARG_NAMED,
     'mypy_extensions.DefaultNamedArg': ARG_NAMED_OPT,
     'mypy_extensions.VarArg': ARG_STAR,
     'mypy_extensions.KwArg': ARG_STAR2,
-}  # type: Final
+}
 
-GENERIC_STUB_NOT_AT_RUNTIME_TYPES = {
+GENERIC_STUB_NOT_AT_RUNTIME_TYPES: Final = {
     'queue.Queue',
     'builtins._PathLike',
     'asyncio.futures.Future',
-}  # type: Final
+}
 
 
 def analyze_type_alias(node: Expression,
@@ -69,7 +70,7 @@ def analyze_type_alias(node: Expression,
                        plugin: Plugin,
                        options: Options,
                        is_typeshed_stub: bool,
-                       allow_unnormalized: bool = False,
+                       allow_new_syntax: bool = False,
                        allow_placeholder: bool = False,
                        in_dynamic_func: bool = False,
                        global_scope: bool = True) -> Optional[Tuple[Type, Set[str]]]:
@@ -80,12 +81,12 @@ def analyze_type_alias(node: Expression,
     Return None otherwise. 'node' must have been semantically analyzed.
     """
     try:
-        type = expr_to_unanalyzed_type(node)
+        type = expr_to_unanalyzed_type(node, options, allow_new_syntax)
     except TypeTranslationError:
         api.fail('Invalid type alias: expression is not a valid type', node)
         return None
     analyzer = TypeAnalyser(api, tvar_scope, plugin, options, is_typeshed_stub,
-                            allow_unnormalized=allow_unnormalized, defining_alias=True,
+                            allow_new_syntax=allow_new_syntax, defining_alias=True,
                             allow_placeholder=allow_placeholder)
     analyzer.in_dynamic_func = in_dynamic_func
     analyzer.global_scope = global_scope
@@ -114,9 +115,9 @@ class TypeAnalyser(SyntheticTypeVisitor[Type], TypeAnalyzerPluginInterface):
     """
 
     # Is this called from an untyped function definition?
-    in_dynamic_func = False  # type: bool
+    in_dynamic_func: bool = False
     # Is this called from global scope?
-    global_scope = True  # type: bool
+    global_scope: bool = True
 
     def __init__(self,
                  api: SemanticAnalyzerCoreInterface,
@@ -126,9 +127,10 @@ class TypeAnalyser(SyntheticTypeVisitor[Type], TypeAnalyzerPluginInterface):
                  is_typeshed_stub: bool, *,
                  defining_alias: bool = False,
                  allow_tuple_literal: bool = False,
-                 allow_unnormalized: bool = False,
+                 allow_new_syntax: bool = False,
                  allow_unbound_tvars: bool = False,
                  allow_placeholder: bool = False,
+                 allow_required: bool = False,
                  report_invalid_types: bool = True) -> None:
         self.api = api
         self.lookup_qualified = api.lookup_qualified
@@ -141,13 +143,15 @@ class TypeAnalyser(SyntheticTypeVisitor[Type], TypeAnalyzerPluginInterface):
         self.allow_tuple_literal = allow_tuple_literal
         # Positive if we are analyzing arguments of another (outer) type
         self.nesting_level = 0
-        # Should we allow unnormalized types like `list[int]`
-        # (currently allowed in stubs)?
-        self.allow_unnormalized = allow_unnormalized
+        # Should we allow new type syntax when targeting older Python versions
+        # like 'list[int]' or 'X | Y' (allowed in stubs)?
+        self.allow_new_syntax = allow_new_syntax
         # Should we accept unbound type variables (always OK in aliases)?
         self.allow_unbound_tvars = allow_unbound_tvars or defining_alias
         # If false, record incomplete ref if we generate PlaceholderType.
         self.allow_placeholder = allow_placeholder
+        # Are we in a context where Required[] is allowed?
+        self.allow_required = allow_required
         # Should we report an error whenever we encounter a RawExpressionType outside
         # of a Literal context: e.g. whenever we encounter an invalid type? Normally,
         # we want to report an error, but the caller may want to do more specialized
@@ -157,7 +161,7 @@ class TypeAnalyser(SyntheticTypeVisitor[Type], TypeAnalyzerPluginInterface):
         self.options = options
         self.is_typeshed_stub = is_typeshed_stub
         # Names of type aliases encountered while analysing a type will be collected here.
-        self.aliases_used = set()  # type: Set[str]
+        self.aliases_used: Set[str] = set()
 
     def visit_unbound_type(self, t: UnboundType, defining_literal: bool = False) -> Type:
         typ = self.visit_unbound_type_nonoptional(t, defining_literal)
@@ -199,7 +203,7 @@ class TypeAnalyser(SyntheticTypeVisitor[Type], TypeAnalyzerPluginInterface):
                 return hook(AnalyzeTypeContext(t, t, self))
             if (fullname in get_nongen_builtins(self.options.python_version)
                     and t.args and
-                    not self.allow_unnormalized and
+                    not self.allow_new_syntax and
                     not self.api.is_future_flag_set("annotations")):
                 self.fail(no_subscript_builtin_alias(fullname,
                                                      propose_alt=not self.defining_alias), t)
@@ -208,22 +212,27 @@ class TypeAnalyser(SyntheticTypeVisitor[Type], TypeAnalyzerPluginInterface):
                 if tvar_def is None:
                     self.fail('ParamSpec "{}" is unbound'.format(t.name), t)
                     return AnyType(TypeOfAny.from_error)
-                self.fail('Invalid location for ParamSpec "{}"'.format(t.name), t)
-                self.note(
-                    'You can use ParamSpec as the first argument to Callable, e.g., '
-                    "'Callable[{}, int]'".format(t.name),
-                    t
+                assert isinstance(tvar_def, ParamSpecType)
+                if len(t.args) > 0:
+                    self.fail('ParamSpec "{}" used with arguments'.format(t.name), t)
+                # Change the line number
+                return ParamSpecType(
+                    tvar_def.name, tvar_def.fullname, tvar_def.id, tvar_def.flavor,
+                    tvar_def.upper_bound, line=t.line, column=t.column,
                 )
-                return AnyType(TypeOfAny.from_error)
             if isinstance(sym.node, TypeVarExpr) and tvar_def is not None and self.defining_alias:
                 self.fail('Can\'t use bound type variable "{}"'
                           ' to define generic alias'.format(t.name), t)
                 return AnyType(TypeOfAny.from_error)
             if isinstance(sym.node, TypeVarExpr) and tvar_def is not None:
-                assert isinstance(tvar_def, TypeVarDef)
+                assert isinstance(tvar_def, TypeVarType)
                 if len(t.args) > 0:
                     self.fail('Type variable "{}" used with arguments'.format(t.name), t)
-                return TypeVarType(tvar_def, t.line)
+                # Change the line number
+                return TypeVarType(
+                    tvar_def.name, tvar_def.fullname, tvar_def.id, tvar_def.values,
+                    tvar_def.upper_bound, tvar_def.variance, line=t.line, column=t.column,
+                )
             special = self.try_analyze_special_unbound_type(t, fullname)
             if special is not None:
                 return special
@@ -247,9 +256,14 @@ class TypeAnalyser(SyntheticTypeVisitor[Type], TypeAnalyzerPluginInterface):
                         python_version=self.options.python_version,
                         use_generic_error=True,
                         unexpanded_type=t)
+                if node.eager:
+                    # TODO: Generate error if recursive (once we have recursive types)
+                    res = get_proper_type(res)
                 return res
             elif isinstance(node, TypeInfo):
                 return self.analyze_type_with_type_info(node, t.args, t)
+            elif node.fullname in ("typing_extensions.TypeAlias", "typing.TypeAlias"):
+                return AnyType(TypeOfAny.special_form)
             else:
                 return self.analyze_unbound_type_without_type_info(t, sym, defining_literal)
         else:  # sym is None
@@ -278,7 +292,8 @@ class TypeAnalyser(SyntheticTypeVisitor[Type], TypeAnalyzerPluginInterface):
             return AnyType(TypeOfAny.from_error)
         elif (fullname == 'typing.Tuple' or
              (fullname == 'builtins.tuple' and (self.options.python_version >= (3, 9) or
-                                                self.api.is_future_flag_set('annotations')))):
+                                                self.api.is_future_flag_set('annotations') or
+                                                self.allow_new_syntax))):
             # Tuple is special because it is involved in builtin import cycle
             # and may be not ready when used.
             sym = self.api.lookup_fully_qualified_or_none('builtins.tuple')
@@ -286,7 +301,7 @@ class TypeAnalyser(SyntheticTypeVisitor[Type], TypeAnalyzerPluginInterface):
                 if self.api.is_incomplete_namespace('builtins'):
                     self.api.record_incomplete_ref()
                 else:
-                    self.fail("Name 'tuple' is not defined", t)
+                    self.fail('Name "tuple" is not defined', t)
                 return AnyType(TypeOfAny.special_form)
             if len(t.args) == 0 and not t.empty_tuple_index:
                 # Bare 'Tuple' is same as 'tuple'
@@ -321,8 +336,8 @@ class TypeAnalyser(SyntheticTypeVisitor[Type], TypeAnalyzerPluginInterface):
                     # To prevent assignment of 'builtins.type' inferred as 'builtins.object'
                     # See https://github.com/python/mypy/issues/9476 for more information
                     return None
-            type_str = 'Type[...]' if fullname == 'typing.Type' else 'type[...]'
             if len(t.args) != 1:
+                type_str = 'Type[...]' if fullname == 'typing.Type' else 'type[...]'
                 self.fail(type_str + ' must have exactly one type argument', t)
             item = self.anal_type(t.args[0])
             return TypeType.make_normalized(item, line=t.line)
@@ -345,6 +360,25 @@ class TypeAnalyser(SyntheticTypeVisitor[Type], TypeAnalyzerPluginInterface):
                           " and at least one annotation", t)
                 return AnyType(TypeOfAny.from_error)
             return self.anal_type(t.args[0])
+        elif fullname in ('typing_extensions.Required', 'typing.Required'):
+            if not self.allow_required:
+                self.fail("Required[] can be only used in a TypedDict definition", t)
+                return AnyType(TypeOfAny.from_error)
+            if len(t.args) != 1:
+                self.fail("Required[] must have exactly one type argument", t)
+                return AnyType(TypeOfAny.from_error)
+            return RequiredType(self.anal_type(t.args[0]), required=True)
+        elif fullname in ('typing_extensions.NotRequired', 'typing.NotRequired'):
+            if not self.allow_required:
+                self.fail("NotRequired[] can be only used in a TypedDict definition", t)
+                return AnyType(TypeOfAny.from_error)
+            if len(t.args) != 1:
+                self.fail("NotRequired[] must have exactly one type argument", t)
+                return AnyType(TypeOfAny.from_error)
+            return RequiredType(self.anal_type(t.args[0]), required=False)
+        elif self.anal_type_guard_arg(t, fullname) is not None:
+            # In most contexts, TypeGuard[...] acts as an alias for bool (ignoring its args)
+            return self.named_type('builtins.bool')
         return None
 
     def get_omitted_any(self, typ: Type, fullname: Optional[str] = None) -> AnyType:
@@ -367,7 +401,8 @@ class TypeAnalyser(SyntheticTypeVisitor[Type], TypeAnalyzerPluginInterface):
         # checked only later, since we do not always know the
         # valid count at this point. Thus we may construct an
         # Instance with an invalid number of type arguments.
-        instance = Instance(info, self.anal_array(args), ctx.line, ctx.column)
+        instance = Instance(info, self.anal_array(args, allow_param_spec=True),
+                            ctx.line, ctx.column)
         # Check type argument count.
         if len(instance.args) != len(info.type_vars) and not self.defining_alias:
             fix_instance(instance, self.fail, self.note,
@@ -452,10 +487,10 @@ class TypeAnalyser(SyntheticTypeVisitor[Type], TypeAnalyzerPluginInterface):
         # to make sure there are no remaining semanal-only types, then give up.
         t = t.copy_modified(args=self.anal_array(t.args))
         # TODO: Move this message building logic to messages.py.
-        notes = []  # type: List[str]
+        notes: List[str] = []
         if isinstance(sym.node, Var):
             notes.append('See https://mypy.readthedocs.io/en/'
-                         'latest/common_issues.html#variables-vs-type-aliases')
+                         'stable/common_issues.html#variables-vs-type-aliases')
             message = 'Variable "{}" is not valid as a type'
         elif isinstance(sym.node, (SYMBOL_FUNCBASE_TYPES, Decorator)):
             message = 'Function "{}" is not valid as a type'
@@ -517,6 +552,9 @@ class TypeAnalyser(SyntheticTypeVisitor[Type], TypeAnalyzerPluginInterface):
     def visit_type_var(self, t: TypeVarType) -> Type:
         return t
 
+    def visit_param_spec(self, t: ParamSpecType) -> Type:
+        return t
+
     def visit_callable_type(self, t: CallableType, nested: bool = True) -> Type:
         # Every Callable can bind its own type variables, if they're not in the outer scope
         with self.tvar_scope_frame():
@@ -524,15 +562,62 @@ class TypeAnalyser(SyntheticTypeVisitor[Type], TypeAnalyzerPluginInterface):
                 variables = t.variables
             else:
                 variables = self.bind_function_type_variables(t, t)
-            ret = t.copy_modified(arg_types=self.anal_array(t.arg_types, nested=nested),
+            special = self.anal_type_guard(t.ret_type)
+            arg_kinds = t.arg_kinds
+            if len(arg_kinds) >= 2 and arg_kinds[-2] == ARG_STAR and arg_kinds[-1] == ARG_STAR2:
+                arg_types = self.anal_array(t.arg_types[:-2], nested=nested) + [
+                    self.anal_star_arg_type(t.arg_types[-2], ARG_STAR, nested=nested),
+                    self.anal_star_arg_type(t.arg_types[-1], ARG_STAR2, nested=nested),
+                ]
+            else:
+                arg_types = self.anal_array(t.arg_types, nested=nested)
+            ret = t.copy_modified(arg_types=arg_types,
                                   ret_type=self.anal_type(t.ret_type, nested=nested),
                                   # If the fallback isn't filled in yet,
                                   # its type will be the falsey FakeInfo
                                   fallback=(t.fallback if t.fallback.type
                                             else self.named_type('builtins.function')),
-                                  variables=self.anal_var_defs(variables))
+                                  variables=self.anal_var_defs(variables),
+                                  type_guard=special,
+                                  )
         return ret
 
+    def anal_type_guard(self, t: Type) -> Optional[Type]:
+        if isinstance(t, UnboundType):
+            sym = self.lookup_qualified(t.name, t)
+            if sym is not None and sym.node is not None:
+                return self.anal_type_guard_arg(t, sym.node.fullname)
+        # TODO: What if it's an Instance? Then use t.type.fullname?
+        return None
+
+    def anal_type_guard_arg(self, t: UnboundType, fullname: str) -> Optional[Type]:
+        if fullname in ('typing_extensions.TypeGuard', 'typing.TypeGuard'):
+            if len(t.args) != 1:
+                self.fail("TypeGuard must have exactly one type argument", t)
+                return AnyType(TypeOfAny.from_error)
+            return self.anal_type(t.args[0])
+        return None
+
+    def anal_star_arg_type(self, t: Type, kind: ArgKind, nested: bool) -> Type:
+        """Analyze signature argument type for *args and **kwargs argument."""
+        # TODO: Check that suffix and kind match
+        if isinstance(t, UnboundType) and t.name and '.' in t.name and not t.args:
+            components = t.name.split('.')
+            sym = self.lookup_qualified('.'.join(components[:-1]), t)
+            if sym is not None and isinstance(sym.node, ParamSpecExpr):
+                tvar_def = self.tvar_scope.get_binding(sym)
+                if isinstance(tvar_def, ParamSpecType):
+                    if kind == ARG_STAR:
+                        flavor = ParamSpecFlavor.ARGS
+                    elif kind == ARG_STAR2:
+                        flavor = ParamSpecFlavor.KWARGS
+                    else:
+                        assert False, kind
+                    return ParamSpecType(tvar_def.name, tvar_def.fullname, tvar_def.id, flavor,
+                                         upper_bound=self.named_type('builtins.object'),
+                                         line=t.line, column=t.column)
+        return self.anal_type(t, nested=nested)
+
     def visit_overloaded(self, t: Overloaded) -> Type:
         # Overloaded types are manually constructed in semanal.py by analyzing the
         # AST and combining together the Callable types this visitor converts.
@@ -629,7 +714,7 @@ class TypeAnalyser(SyntheticTypeVisitor[Type], TypeAnalyzerPluginInterface):
         assert False, "Internal error: Unexpected partial type"
 
     def visit_ellipsis_type(self, t: EllipsisType) -> Type:
-        self.fail("Unexpected '...'", t)
+        self.fail('Unexpected "..."', t)
         return AnyType(TypeOfAny.from_error)
 
     def visit_type_type(self, t: TypeType) -> Type:
@@ -658,17 +743,20 @@ class TypeAnalyser(SyntheticTypeVisitor[Type], TypeAnalyzerPluginInterface):
         if sym is None:
             return None
         tvar_def = self.tvar_scope.get_binding(sym)
-        if not isinstance(tvar_def, ParamSpecDef):
+        if not isinstance(tvar_def, ParamSpecType):
             return None
 
-        # TODO(shantanu): construct correct type for paramspec
+        # TODO: Use tuple[...] or Mapping[..] instead?
+        obj = self.named_type('builtins.object')
         return CallableType(
-            [AnyType(TypeOfAny.explicit), AnyType(TypeOfAny.explicit)],
+            [ParamSpecType(tvar_def.name, tvar_def.fullname, tvar_def.id, ParamSpecFlavor.ARGS,
+                           upper_bound=obj),
+             ParamSpecType(tvar_def.name, tvar_def.fullname, tvar_def.id, ParamSpecFlavor.KWARGS,
+                           upper_bound=obj)],
             [nodes.ARG_STAR, nodes.ARG_STAR2],
             [None, None],
             ret_type=ret_type,
             fallback=fallback,
-            is_ellipsis_args=True
         )
 
     def analyze_callable_type(self, t: UnboundType) -> Type:
@@ -676,12 +764,7 @@ class TypeAnalyser(SyntheticTypeVisitor[Type], TypeAnalyzerPluginInterface):
         if len(t.args) == 0:
             # Callable (bare). Treat as Callable[..., Any].
             any_type = self.get_omitted_any(t)
-            ret = CallableType([any_type, any_type],
-                               [nodes.ARG_STAR, nodes.ARG_STAR2],
-                               [None, None],
-                               ret_type=any_type,
-                               fallback=fallback,
-                               is_ellipsis_args=True)
+            ret = callable_with_ellipsis(any_type, any_type, fallback)
         elif len(t.args) == 2:
             callable_args = t.args[0]
             ret_type = t.args[1]
@@ -698,13 +781,9 @@ class TypeAnalyser(SyntheticTypeVisitor[Type], TypeAnalyzerPluginInterface):
                                    fallback=fallback)
             elif isinstance(callable_args, EllipsisType):
                 # Callable[..., RET] (with literal ellipsis; accept arbitrary arguments)
-                ret = CallableType([AnyType(TypeOfAny.explicit),
-                                    AnyType(TypeOfAny.explicit)],
-                                   [nodes.ARG_STAR, nodes.ARG_STAR2],
-                                   [None, None],
-                                   ret_type=ret_type,
-                                   fallback=fallback,
-                                   is_ellipsis_args=True)
+                ret = callable_with_ellipsis(AnyType(TypeOfAny.explicit),
+                                             ret_type=ret_type,
+                                             fallback=fallback)
             else:
                 # Callable[P, RET] (where P is ParamSpec)
                 maybe_ret = self.analyze_callable_args_for_paramspec(
@@ -714,7 +793,7 @@ class TypeAnalyser(SyntheticTypeVisitor[Type], TypeAnalyzerPluginInterface):
                 )
                 if maybe_ret is None:
                     # Callable[?, RET] (where ? is something invalid)
-                    # TODO(shantanu): change error to mention paramspec, once we actually have some
+                    # TODO(PEP612): change error to mention paramspec, once we actually have some
                     # support for it
                     self.fail('The first argument to Callable must be a list of types or "..."', t)
                     return AnyType(TypeOfAny.from_error)
@@ -726,11 +805,11 @@ class TypeAnalyser(SyntheticTypeVisitor[Type], TypeAnalyzerPluginInterface):
         return ret.accept(self)
 
     def analyze_callable_args(self, arglist: TypeList) -> Optional[Tuple[List[Type],
-                                                                         List[int],
+                                                                         List[ArgKind],
                                                                          List[Optional[str]]]]:
-        args = []   # type: List[Type]
-        kinds = []  # type: List[int]
-        names = []  # type: List[Optional[str]]
+        args: List[Type] = []
+        kinds: List[ArgKind] = []
+        names: List[Optional[str]] = []
         for arg in arglist.items:
             if isinstance(arg, CallableArgument):
                 args.append(arg.typ)
@@ -749,7 +828,7 @@ class TypeAnalyser(SyntheticTypeVisitor[Type], TypeAnalyzerPluginInterface):
                     assert found.fullname is not None
                     kind = ARG_KINDS_BY_CONSTRUCTOR[found.fullname]
                     kinds.append(kind)
-                    if arg.name is not None and kind in {ARG_STAR, ARG_STAR2}:
+                    if arg.name is not None and kind.is_star():
                         self.fail("{} arguments should not have names".format(
                             arg.constructor), arg)
                         return None
@@ -767,7 +846,7 @@ class TypeAnalyser(SyntheticTypeVisitor[Type], TypeAnalyzerPluginInterface):
             self.fail('Literal[...] must have at least one parameter', t)
             return AnyType(TypeOfAny.from_error)
 
-        output = []  # type: List[Type]
+        output: List[Type] = []
         for i, arg in enumerate(t.args):
             analyzed_types = self.analyze_literal_param(i + 1, arg, t)
             if analyzed_types is None:
@@ -804,7 +883,7 @@ class TypeAnalyser(SyntheticTypeVisitor[Type], TypeAnalyzerPluginInterface):
             #
             # 1. If the user attempts use an explicit Any as a parameter
             # 2. If the user is trying to use an enum value imported from a module with
-            #    no type hints, giving it an an implicit type of 'Any'
+            #    no type hints, giving it an implicit type of 'Any'
             # 3. If there's some other underlying problem with the parameter.
             #
             # We report an error in only the first two cases. In the third case, we assume
@@ -869,8 +948,8 @@ class TypeAnalyser(SyntheticTypeVisitor[Type], TypeAnalyzerPluginInterface):
     def infer_type_variables(self,
                              type: CallableType) -> List[Tuple[str, TypeVarLikeExpr]]:
         """Return list of unique type variables referred to in a callable."""
-        names = []  # type: List[str]
-        tvars = []  # type: List[TypeVarLikeExpr]
+        names: List[str] = []
+        tvars: List[TypeVarLikeExpr] = []
         for arg in type.arg_types:
             for name, tvar_expr in arg.accept(
                 TypeVarLikeQuery(self.lookup_qualified, self.tvar_scope)
@@ -892,7 +971,7 @@ class TypeAnalyser(SyntheticTypeVisitor[Type], TypeAnalyzerPluginInterface):
 
     def bind_function_type_variables(
         self, fun_type: CallableType, defn: Context
-    ) -> Sequence[TypeVarLikeDef]:
+    ) -> Sequence[TypeVarLikeType]:
         """Find the type variables of the function type and bind them in our tvar_scope"""
         if fun_type.variables:
             for var in fun_type.variables:
@@ -906,10 +985,10 @@ class TypeAnalyser(SyntheticTypeVisitor[Type], TypeAnalyzerPluginInterface):
         # Do not define a new type variable if already defined in scope.
         typevars = [(name, tvar) for name, tvar in typevars
                     if not self.is_defined_type_var(name, defn)]
-        defs = []  # type: List[TypeVarLikeDef]
+        defs: List[TypeVarLikeType] = []
         for name, tvar in typevars:
             if not self.tvar_scope.allow_binding(tvar.fullname):
-                self.fail("Type variable '{}' is bound by an outer class".format(name), defn)
+                self.fail('Type variable "{}" is bound by an outer class'.format(name), defn)
             self.tvar_scope.bind_new(name, tvar)
             binding = self.tvar_scope.get_binding(tvar.fullname)
             assert binding is not None
@@ -923,24 +1002,40 @@ class TypeAnalyser(SyntheticTypeVisitor[Type], TypeAnalyzerPluginInterface):
             return False
         return self.tvar_scope.get_binding(tvar_node) is not None
 
-    def anal_array(self, a: Iterable[Type], nested: bool = True) -> List[Type]:
-        res = []  # type: List[Type]
+    def anal_array(self,
+                   a: Iterable[Type],
+                   nested: bool = True, *,
+                   allow_param_spec: bool = False) -> List[Type]:
+        res: List[Type] = []
         for t in a:
-            res.append(self.anal_type(t, nested))
+            res.append(self.anal_type(t, nested, allow_param_spec=allow_param_spec))
         return res
 
-    def anal_type(self, t: Type, nested: bool = True) -> Type:
+    def anal_type(self, t: Type, nested: bool = True, *, allow_param_spec: bool = False) -> Type:
         if nested:
             self.nesting_level += 1
+        old_allow_required = self.allow_required
+        self.allow_required = False
         try:
-            return t.accept(self)
+            analyzed = t.accept(self)
         finally:
             if nested:
                 self.nesting_level -= 1
+            self.allow_required = old_allow_required
+        if (not allow_param_spec
+                and isinstance(analyzed, ParamSpecType)
+                and analyzed.flavor == ParamSpecFlavor.BARE):
+            self.fail('Invalid location for ParamSpec "{}"'.format(analyzed.name), t)
+            self.note(
+                'You can use ParamSpec as the first argument to Callable, e.g., '
+                "'Callable[{}, int]'".format(analyzed.name),
+                t
+            )
+        return analyzed
 
-    def anal_var_def(self, var_def: TypeVarLikeDef) -> TypeVarLikeDef:
-        if isinstance(var_def, TypeVarDef):
-            return TypeVarDef(
+    def anal_var_def(self, var_def: TypeVarLikeType) -> TypeVarLikeType:
+        if isinstance(var_def, TypeVarType):
+            return TypeVarType(
                 var_def.name,
                 var_def.fullname,
                 var_def.id.raw_id,
@@ -952,7 +1047,7 @@ class TypeAnalyser(SyntheticTypeVisitor[Type], TypeAnalyzerPluginInterface):
         else:
             return var_def
 
-    def anal_var_defs(self, var_defs: Sequence[TypeVarLikeDef]) -> List[TypeVarLikeDef]:
+    def anal_var_defs(self, var_defs: Sequence[TypeVarLikeType]) -> List[TypeVarLikeType]:
         return [self.anal_var_def(vd) for vd in var_defs]
 
     def named_type_with_normalized_str(self, fully_qualified_name: str) -> Instance:
@@ -1018,8 +1113,8 @@ def get_omitted_any(disallow_any: bool, fail: MsgCallback, note: MsgCallback,
                 # (string literal escaping) for classes not generic at runtime
                 note(
                     "Subscripting classes that are not generic at runtime may require "
-                    "escaping, see https://mypy.readthedocs.io/"
-                    "en/latest/common_issues.html#not-generic-runtime",
+                    "escaping, see https://mypy.readthedocs.io/en/stable/runtime_troubles.html"
+                    "#not-generic-runtime",
                     typ,
                     code=codes.TYPE_ARG)
 
@@ -1041,7 +1136,7 @@ def fix_instance(t: Instance, fail: MsgCallback, note: MsgCallback,
     """
     if len(t.args) == 0:
         if use_generic_error:
-            fullname = None  # type: Optional[str]
+            fullname: Optional[str] = None
         else:
             fullname = t.type.fullname
         any_type = get_omitted_any(disallow_any, fail, note, t, python_version, fullname,
@@ -1140,8 +1235,8 @@ def set_any_tvars(node: TypeAlias,
 
 def remove_dups(tvars: Iterable[T]) -> List[T]:
     # Get unique elements in order of appearance
-    all_tvars = set()  # type: Set[T]
-    new_tvars = []  # type: List[T]
+    all_tvars: Set[T] = set()
+    new_tvars: List[T] = []
     for t in tvars:
         if t not in all_tvars:
             new_tvars.append(t)
@@ -1154,6 +1249,7 @@ def flatten_tvars(ll: Iterable[List[T]]) -> List[T]:
 
 
 class TypeVarLikeQuery(TypeQuery[TypeVarLikeList]):
+    """Find TypeVar and ParamSpec references in an unbound type."""
 
     def __init__(self,
                  lookup: Callable[[str, Context], Optional[SymbolTableNode]],
@@ -1176,7 +1272,17 @@ class TypeVarLikeQuery(TypeQuery[TypeVarLikeList]):
 
     def visit_unbound_type(self, t: UnboundType) -> TypeVarLikeList:
         name = t.name
-        node = self.lookup(name, t)
+        node = None
+        # Special case P.args and P.kwargs for ParamSpecs only.
+        if name.endswith('args'):
+            if name.endswith('.args') or name.endswith('.kwargs'):
+                base = '.'.join(name.split('.')[:-1])
+                n = self.lookup(base, t)
+                if n is not None and isinstance(n.node, ParamSpecExpr):
+                    node = n
+                    name = base
+        if node is None:
+            node = self.lookup(name, t)
         if node and isinstance(node.node, TypeVarLikeExpr) and (
                 self.include_bound_tvars or self.scope.get_binding(node) is None):
             assert isinstance(node.node, TypeVarLikeExpr)
@@ -1185,6 +1291,9 @@ class TypeVarLikeQuery(TypeQuery[TypeVarLikeList]):
             return []
         elif node and node.fullname in ('typing_extensions.Literal', 'typing.Literal'):
             return []
+        elif node and node.fullname in ('typing_extensions.Annotated', 'typing.Annotated'):
+            # Don't query the second argument to Annotated for TypeVars
+            return self.query_types([t.args[0]])
         else:
             return super().visit_unbound_type(t)
 
@@ -1261,7 +1370,7 @@ class CollectAnyTypesQuery(TypeQuery[List[AnyType]]):
 
     @classmethod
     def combine_lists_strategy(cls, it: Iterable[List[AnyType]]) -> List[AnyType]:
-        result = []  # type: List[AnyType]
+        result: List[AnyType] = []
         for l in it:
             result.extend(l)
         return result
