@@ -30,14 +30,14 @@ DEFAULT_COLUMNS: Final = 80
 
 # At least this number of columns will be shown on each side of
 # error location when printing source code snippet.
-MINIMUM_WIDTH = 20
+MINIMUM_WIDTH: Final = 20
 
 # VT100 color code processing was added in Windows 10, but only the second major update,
 # Threshold 2. Fortunately, everyone (even on LTSB, Long Term Support Branch) should
 # have a version of Windows 10 newer than this. Note that Windows 8 and below are not
 # supported, but are either going out of support, or make up only a few % of the market.
-MINIMUM_WINDOWS_MAJOR_VT100 = 10
-MINIMUM_WINDOWS_BUILD_VT100 = 10586
+MINIMUM_WINDOWS_MAJOR_VT100: Final = 10
+MINIMUM_WINDOWS_BUILD_VT100: Final = 10586
 
 default_python2_interpreter: Final = [
     "python2",
@@ -566,11 +566,12 @@ class FancyFormatter:
         under = curses.tigetstr('smul')
         set_color = curses.tigetstr('setaf')
         set_eseq = curses.tigetstr('cup')
+        normal = curses.tigetstr('sgr0')
 
-        if not (bold and under and set_color and set_eseq):
+        if not (bold and under and set_color and set_eseq and normal):
             return False
 
-        self.NORMAL = curses.tigetstr('sgr0').decode()
+        self.NORMAL = normal.decode()
         self.BOLD = bold.decode()
         self.UNDER = under.decode()
         self.DIM = parse_gray_color(set_eseq)
